local AceOO = AceLibrary("AceOO-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")


if not AceLibrary:HasInstance("PratPlugin-1.0") then
    DreamChat.FrameClass = AceOO.Class("AceEvent-2.0")
else
    DreamChat.FrameClass = AceOO.Class("AceEvent-2.0","AceHook-2.1", "PratPlugin-1.0")
end    

-- @desc Initializer function for a DreamChat frame.
-- @arg id string The identifier for the frame.
-- @arg desc string The description for the frame. (Shown name?)
function DreamChat.FrameClass.prototype:init(id, desc)
	DreamChat.FrameClass.super.prototype.init(self)
	
	self.id = id
	self.name = "DreamChat"..id
	self.db = DreamChat:AcquireDBNamespace(self.name)
	DreamChat:RegisterDefaults(self.name, "profile", { 
		color = { r = 0, g = 0, b = 0},
		bgAlpha = 0.25,
		alpha=1,
		minimized = false,
		height = 128,
		width = 430,
		PosX = 250,
		PosY = 250,
		acceptedTypes={},
		fontSize=12,
		showTitle=false,
		title=(desc or "Untitled"),
		attachToFrame=nil,
		attachPoint=nil,
		attachTo=nil,
		attachX=0,
		attachY=0
	} )

	self:CreateFrame()
end

-- @desc Creates the WoW UI frame for this DreamChat frame.
function DreamChat.FrameClass.prototype:CreateFrame() 
	self.frame = CreateFrame("ScrollingMessageFrame", self.name, UIParent, "DreamChatTemplate")
	self.frame.class = self
	
	self.Background = getglobal(self.name.."Background")
	self.Title = getglobal(self.name.."Title")
	self.Button = getglobal(self.name.."ToggleButton")
	self.Button:RegisterForClicks("LeftButtonUp","RightButtonUp")
	self.Resize = {}
	self.Resize.TopLeft = getglobal(self.name.."ResizeTopLeftTexture")
	self.Resize.TopRight = getglobal(self.name.."ResizeTopRightTexture")
	self.Resize.BottomLeft = getglobal(self.name.."ResizeBottomLeftTexture")
	self.Resize.BottomRight = getglobal(self.name.."ResizeBottomRightTexture")
	self.Resize.Top = getglobal(self.name.."ResizeTopTexture")
	self.Resize.Bottom = getglobal(self.name.."ResizeBottomTexture")
	self.Resize.Left = getglobal(self.name.."ResizeLeftTexture")
	self.Resize.Right = getglobal(self.name.."ResizeRightTexture")
	
	self:UpdateLook()
	self.frame:Show()
	self:RefreshLayout()
	self:CheckButtons()
	
end

-- @desc Checks the current status of window and updates the correct button.
function DreamChat.FrameClass.prototype:CheckButtons()
	if self.db.profile.minimized then
		self.Button:SetNormalTexture("Interface\\AddOns\\DreamChat\\textures\\button-up")
		self.Button:SetPushedTexture("Interface\\AddOns\\DreamChat\\textures\\button-up-pushed")
	else
		self.Button:SetNormalTexture("Interface\\AddOns\\DreamChat\\textures\\button-down")
		self.Button:SetPushedTexture("Interface\\AddOns\\DreamChat\\textures\\button-down-pushed")
	end
end

-- @desc Updates some look and feel properties of the frame. Currently font size, title and background color.
function DreamChat.FrameClass.prototype:UpdateLook()
	-- Todo!: Changing the color doesn't work; have to change to a white background... vertex color is modulative
	self.Background:SetVertexColor(self.db.profile.color.r, self.db.profile.color.g, self.db.profile.color.b)
	self.Background:SetAlpha(self.db.profile.bgAlpha)
	for _,frame in pairs(self.Resize) do
		frame:SetVertexColor(self.db.profile.color.r, self.db.profile.color.g, self.db.profile.color.b)
	end
	local fn,fs,ff;
	fn,fs,ff=self.frame:GetFont()
	self.frame:SetFont(fn,self.db.profile.fontSize,ff)
	self.frame:SetAlpha(self.db.profile.alpha)
	if self.db.profile.showTitle then
		self.Title:SetText(self.db.profile.title)
		self.Title:Show()
	else
		self.Title:Hide()
	end
end

-- @desc Refreshes the layout of a frame (position and size)
function DreamChat.FrameClass.prototype:RefreshLayout()
	self:LoadPosition()
	self:LoadSize()
	self:LoadPosition() -- Just in case coding. My bad habit.
end

-- @desc Loads the frame position from the database (called by RefreshLayout)
function DreamChat.FrameClass.prototype:LoadPosition()
	local x,y,s = self.db.profile.PosX, self.db.profile.PosY, self.frame:GetEffectiveScale()
	x,y = x/s, y/s
	self.frame:ClearAllPoints()
	if self.db.profile.attachToFrame~=nil then
		self.frame:SetPoint(self.db.profile.attachPoint, getglobal(self.db.profile.attachToFrame), self.db.profile.attachTo, self.db.profile.attachX, self.db.profile.attachY)
	else
		self.frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x, y)
	end
end

-- @desc Loads the frame size (called by RefreshLayout)
function DreamChat.FrameClass.prototype:LoadSize()
	self.frame:SetWidth(self.db.profile.width)
	self.frame:SetHeight(self:GetFrameHeight())
end

-- @desc Saves the frame position.
function DreamChat.FrameClass.prototype:SavePosition()
	local x,y = self.frame:GetLeft(), self.frame:GetBottom()
	local s = self.frame:GetEffectiveScale()
	self.db.profile.PosX, self.db.profile.PosY = x*s, y*s
end

-- @desc Saves the frame size.
function DreamChat.FrameClass.prototype:SaveSize()
	self.db.profile.width = self.frame:GetWidth()
	if not self.db.profile.minimized then
		self.db.profile.height = self.frame:GetHeight()
	end
end

-- @desc Called when the user begins moving the frame.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:DragStart()
	if self.db.profile.attachToFrame~=nil then
		-- Todo: Error message when not movable?
	else
		self.frame:StartMoving()
	end
end

-- @desc Called when the user stops moving the frame and saves its information to the database.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:DragStop()
	-- Todo!: Reanchor by bottomleft to prevent leaping.
	self.frame:StopMovingOrSizing()
	self:SavePosition()
end

-- @desc Mouse wheel event handler for the frame.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:MouseWheel()
	if (arg1 > 0) then
		if (IsShiftKeyDown()) then self.frame:ScrollToTop();
		else self.frame:ScrollUp() end
	elseif (arg1 < 0) then
		if (IsShiftKeyDown()) then self.frame:ScrollToBottom();
		else self.frame:ScrollDown() end
	end
end

-- @desc Hyperlink click handler for the frame.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:Hyperlink(arg1,arg2,arg3)
	if ChatFrame_OnHyperlinkShow~=nil then
		ChatFrame_OnHyperlinkShow(arg1,arg2,arg3)
	else
		SetItemRef(arg1,arg2,arg3)
	end
end

-- @desc Called when the user starts resizing the frame from the said anchor. If the frame is minimized, only horizontal resizing is possible.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:StartResize(anchor)
	if self.db.profile.minimized and not ( anchor == "LEFT" or anchor == "RIGHT" ) then return end
	self.frame:StartSizing(anchor)
end

-- @desc Called when the user stops resizing the frame. Saves the new size in the database.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:StopResize()
	self.frame:StopMovingOrSizing()
	self:SaveSize()
	self:SetAttachment()
	
end

-- @desc Returns the "effective" frame height (depending on minimization state) for the frame.
function DreamChat.FrameClass.prototype:GetFrameHeight()
	-- Todo?: If the frame font size is not 14, this should probably be changed. (done)
	if self.db.profile.minimized then
		return self:GetLastFontstringHieght()
		-- This was originally 20.
	else
		return self.db.profile.height
	end
end

-- @desc Called when the user clicks the minimizer button.
-- @cat EventHandler
function DreamChat.FrameClass.prototype:ToggleButtonClick()
	-- Todo: Pass arg1 from event handler?
	if arg1=="LeftButton" then
		self:RefreshLayout()
		self.db.profile.minimized = not self.db.profile.minimized
		if self.db.profile.minimized then
			self.frame:ScrollToBottom()
			self:CheckButtons()
		else
			self:CheckButtons()
		end
		DreamChat:SetFrameHeight(self.frame, self:GetFrameHeight())
	elseif arg1=="RightButton" then
		self:ShowPopupMenu()
	end
end

function DreamChat.FrameClass.prototype:GetLastFontstringHieght()
    local fs = self:GetLastline(self.frame:GetRegions())
    local h = math.max(fs and fs:GetHeight() or 0, self.db.profile.fontSize)+6
    return tostring(h)
end

function DreamChat.FrameClass.prototype:GetLastline(...)
  for i=1,select("#", ...) do
    local x = select(i, ...)
	if x:GetObjectType() == "FontString" and x:GetName() == nil then
		return x
	end
  end
end


-- @desc Shows the popup menu.
function DreamChat.FrameClass.prototype:ShowPopupMenu()
	dewdrop:Open(self.Button,'children',function(l,v,v1,v2,v3,v4)self:FeedDewdrop(l,v,v1,v2,v3,v4)end)
end

-- @desc Feeds Dewdrop the popup menu stuff. Called by Dewdrop via ShowPopupMenu()
function DreamChat.FrameClass.prototype:FeedDewdrop(l,v,vn1,vn2,vn3,vn4)
	-- Todo: Full menu

	if l==1 then
		dewdrop:AddLine('text', 'Message Types...','hasArrow',true,'value','accepts')
		dewdrop:AddLine('text', 'Appearance...','hasArrow',true,'value','appearance')
		dewdrop:AddLine('text', 'Attach...','hasArrow',true,'value','attach')
		dewdrop:AddLine('text', 'Clear','func',function()self.frame:Clear()end)
		dewdrop:AddLine('text', 'Hide','func',function()self.frame:Hide()end)
		dewdrop:AddLine('text', 'New Frame','hasArrow',true,'func',function()DreamChat:CreateChatFrame()end)
		dewdrop:AddLine('text', '(Reset Editbox)','func',function()DreamChat:ResetEditBoxPosition()end)
	elseif l==2 then
		if v=="accepts" then
			self:FeedAcceptsMenu()
		elseif v=="appearance" then
			self:FeedAppearanceMenu()
		elseif v=="attach" then
			self:FeedAttachMenu()
		end
	elseif l==3 then
		if vn1=="accepts" then
			self:FeedAcceptsMenu(v,l)
		elseif vn1=="attach" then
			self:FeedAttachMenu(v)
		end
	elseif l==4 then
		if vn2=="accepts" then
			self:FeedAcceptsMenu(v,l)
		end
	end
end

-- @desc Changes a look setting for this frame.
-- @arg colorId string identifier (typeId, "background", "fontSize" or "alpha", currently)
-- @arg r number red value (for colors)
-- @arg g number green value (for colors)
-- @arg b number blue value (for colors)
-- @arg a number alpha (or other) value
function DreamChat.FrameClass.prototype:UpdateLookProperty(lookId,r,g,b,a)
	-- Todo?: Support the typeID colors.
	if lookId==nil then return; end
	if lookId=="background" then
		-- Keeping garbage collection happier (no {r,g,b})
		self.db.profile.color[1]=r
		self.db.profile.color[2]=g
		self.db.profile.color[3]=b
		self.db.profile.bgAlpha=a
		self:UpdateLook()
	elseif lookId=="alpha" then
		self.db.profile.alpha=a
		self:UpdateLook()
	elseif lookId=="fontSize" then
		self.db.profile.fontSize=a
		self:UpdateLook()
	else
		if DreamChat.db.profile.colors[lookId]==nil then
			DreamChat.db.profile.colors[lookId]={}
		end	
		DreamChat.db.profile.colors[lookId][1]=r
		DreamChat.db.profile.colors[lookId][2]=g
		DreamChat.db.profile.colors[lookId][3]=b
		self:UpdateLook()	
	end
end

-- @desc Sets the frame attachment for this frame. Pass nil to attachToFrame to make the frame floating again.
-- @arg attachToFrame string frame to attach to
-- @arg attachThisEdge string attachment point on THIS frame
-- @arg attachToEdge string attachment point on THAT frame
function DreamChat.FrameClass.prototype:SetAttachment(attachToFrame,attachThisEdge,attachToEdge,offset)
	DreamChat:Debug(attachToFrame,attachThisEdge,attachToEdge)
	if attachToFrame==nil or getglobal(attachToFrame)==nil then
		self.db.profile.attachToFrame=nil
		self.db.profile.attachPoint=nil
		self.db.profile.attachTo=nil
		self.db.profile.attachX,self.db.profile.attachY=nil,nil
	else
		
		self.db.profile.attachToFrame=attachToFrame
		self.db.profile.attachPoint=attachThisEdge
		self.db.profile.attachTo=attachToEdge
		if offset then
			self.db.profile.attachX,self.db.profile.attachY=unpack(offset)
		else
			self.db.profile.attachX,self.db.profile.attachY=0,0
		end
	end
	self:RefreshLayout()
end

-- @desc Feeds Dewdrop the frame accepts menu. Called by FeedDewdrop()
function DreamChat.FrameClass.prototype:FeedAcceptsMenu(val,l)
	local keys,_,msgType,n;
	if val==nil then
		keys={};
		for k,v in pairs(DreamChat.messageCategs) do
			table.insert(keys,k);
		end
		table.sort(keys);
		for k,v in pairs(keys) do
			dewdrop:AddLine('text',v,'value',v,'hasArrow',true)
		end
	else
		if l==3 then
			for _,msgType in ipairs(DreamChat.messageCategs[val]) do
				local r,g,b=DreamChat:GetMessageTypeColor(msgType);
				dewdrop:AddLine(
						'text',DreamChat.messageTypes[msgType].name or msgType,
						'arg1',msgType,
						'value',msgType,
						'checked',self:AcceptsMessageType(msgType),
						'func',function(a)self:ToggleAcceptsMessageType(a)end,
						'hasArrow',true,
						'textR',r,'textG',g,'textB',b)
			end
		else
			msgType=val;
			local r,g,b=DreamChat:GetMessageTypeColor(msgType);
			
			dewdrop:AddLine('text','Color','hasColorSwatch',true,'r',r,'g',g,'b',b,'colorFunc',function(r,g,b)self:UpdateLookProperty(msgType,r,g,b)end)
		end
	end
		
end

local attachmentModes=
	{
	--	Description		AttachTo	AttachPoint	Offset
		{"Left of !",		"RIGHT",	"LEFT",		{-10,0}	},
		{"Right of !",		"LEFT",		"RIGHT",	{10,0}	},
		{"Above !",		"BOTTOM",	"TOP",		{0,10}	},
		{"Below !",		"TOP",		"BOTTOM",	{0,-10}	},
		{"Above (left) !",	"BOTTOMLEFT",	"TOPLEFT",	{0,10}	},
		{"Above (right) !",	"BOTTOMRIGHT",	"TOPRIGHT",	{0,10}	},
		{"Below (left) !",	"TOPLEFT",	"BOTTOMLEFT",	{0,-10}	},
		{"Below (right) !",	"TOPRIGHT",	"BOTTOMRIGHT",	{0,-10} },

	};
		

-- @desc Feeds Dewdrop the frame attachments menu. Called by FeedDewdrop()
function DreamChat.FrameClass.prototype:FeedAttachMenu(val)
	local title,i,_,c,t;	
	if val==nil then
		dewdrop:AddLine('text','Floating','value',v,'checked',(self.db.profile.attachToFrame==nil),'func',function()self:SetAttachment(nil)end)
		for i=1,DreamChat.numframes do
			dewdrop:AddLine('text',string.format("Attach to [%d] %s...",i, DreamChat.frames[i].db.profile.title),'value',i,'hasArrow',true,'disabled',(i==self.id),'checked',(self.db.profile.attachToFrame=='DreamChat'..i))
		end
	else
		i=val; -- Lazy
		title=format("[%d] %s...",i, DreamChat.frames[i].db.profile.title)
		for _,c in ipairs(attachmentModes) do
			t=gsub(c[1],'!',title)
			dewdrop:AddLine(	'text',t,
						'checked',(self.db.profile.attachToFrame=='DreamChat'..i and self.db.profile.attachPoint==c[2] and self.db.profile.attachTo==c[3]),
						'arg1','DreamChat'..i,'arg2',c[2],'arg3',c[3],'arg4',c[4], -- yes.
						'func',function(a,b,c,d)self:SetAttachment(a,b,c,d)end)
		end
		dewdrop:AddLine('text','Match Width','func',
					function()
						self.frame:SetWidth(getglobal('DreamChat'..i):GetWidth())
						self:SaveSize()
						self:RefreshLayout()
					end)
		-- Todo: Revise this line
		dewdrop:AddLine('text','Match Height','func',
					function()
						self.frame:SetHeight(getglobal('DreamChat'..i):GetHeight())
						self:SaveSize()
						self:RefreshLayout()
					end)

	end
		
end

-- @desc Feeds Dewdrop the frame appearance menu. Called by FeedDewdrop()
function DreamChat.FrameClass.prototype:FeedAppearanceMenu(val)
	dewdrop:AddLine(	'text', 'Background',
				'hasColorSwatch',true,
				'hasOpacity',true,
				'r',self.db.profile.color[1],
				'g',self.db.profile.color[2],
				'b',self.db.profile.color[3],
				'opacity',self.db.profile.bgAlpha,
				'colorFunc',function(r,g,b,a)self:UpdateLookProperty("background",r,g,b,a)end)
	dewdrop:AddLine(	'text', 'Opacity',
				'hasSlider',true,
				'hasArrow',true,
				'sliderMin',0.05,
				'sliderMax',1,
				'sliderIsPercent',true,
				'sliderValue',self.db.profile.alpha,
				'sliderFunc',function(v)self:UpdateLookProperty("alpha",nil,nil,nil,v)end)
	dewdrop:AddLine(	'text', 'Font Size',
				'hasSlider',true,
				'hasArrow',true,
				'sliderMin',7,
				'sliderMax',30,
				'sliderValue',self.db.profile.fontSize,
				'sliderStep',0.25,
				'sliderFunc',function(v)self:UpdateLookProperty("fontSize",nil,nil,nil,v)end)
	dewdrop:AddLine(	'text', 'Title',
				'checked',(self.db.profile.showTitle),
				'hasArrow',true,
				'hasEditBox',true,
				'editBoxText',self.db.profile.title,
				'editBoxChangeFunc',function(v)self.db.profile.title=v;self:UpdateLook()end,
				'func',function(v)self.db.profile.showTitle=not self.db.profile.showTitle;self:UpdateLook()end)
end


-- @desc Asks the frame whether it accepts (might print) a message of type typeId.
-- @arg typeId The type ID.
function DreamChat.FrameClass.prototype:AcceptsMessageType(typeId)
	return (self.db.profile.acceptedTypes[typeId]~=nil)
end



DreamChat.FrameClass.prototypeinfo = {}
DreamChat.FrameClass.prototypemsg = {}

-- @desc Shows a message in the DreamChat window.
-- @arg message table The message structure.
-- @ret boolean True if succeeds, nil if not.
-- @note Currently will print every message type.
function DreamChat.FrameClass.prototype:ShowMessage(message)
	local str="";
	local r,g,b,text;
	local sender=message.sender;
	local typeId=message.typeId;
	
	r,g,b=DreamChat:GetMessageTypeColor(typeId);

    if not self.pratFrame and type(self.GetPratFormatProvider) == "function" then 
        -- Register the frame with Prat
        self.pratFrame = self:RegisterExternalFrame(getglobal(self.name))
    end	   

	if DreamChat.messageTypes[typeId].pratRenderable and self.pratFrame then    
	    local m = self:GetExternalMessageTable()
	    
	    local type = DreamChat.MessageSenderMap[typeId]
        if type and self.pratFrame then
            local event = "CHAT_MSG_"..type
            m = self:BuildExternalMessage(m, type, 
                        message.channel, message.channelName,
                        sender, nil, 0,
                        message.text, 
                        r, g, b,
                        message.text)
                     
        	str = self:FormatTextForFrame(self.pratFrame, m, event)
        end
	else
		text=DreamChat:Languagize(message);
		if sender~=nil and sender~="" then
			str=DreamChat:RenderSender(message)..": ";
		end
		if DreamChat.messageTypes[typeId].prefix then
			str=DreamChat.messageTypes[typeId].prefix..str;
		end
		if DreamChat.messageTypes[typeId].dontShowName then
			str=str..(message.text);
		else
			str=DreamChat.messageTypes[typeId].name.." "..str..(message.text);
		end
	end
	
	local ts = ""
	if type(self.GetPratFormatProvider) == "function" then
        ts = self:GetPratFormatProvider("TIMESTAMPS", self.pratFrame) or ""
    end	
    	
    if strlen(str) > 0 then
    	getglobal(self.name):AddMessage(ts..str,r,g,b,1)
	    self:RefreshLayout()
	end
	
	return true
end

-- @desc Sets a single type id acceptance property.
-- @arg typeId string The type ID.
-- @arg accepts boolean Whether this frame accepts the said typeId.
-- @note The type ID needs not exist.
function DreamChat.FrameClass.prototype:SetAcceptsMessageType(typeId,accepts)
	typeId=string.lower(typeId);
	if accepts then
		self.db.profile.acceptedTypes[typeId]=true;
	else
		self.db.profile.acceptedTypes[typeId]=nil;
	end
end

-- @desc Toggles a type id acceptance property. Shorthand for dewdrop use.
-- @arg typeId string The typeId.
function DreamChat.FrameClass.prototype:ToggleAcceptsMessageType(typeId)
	self:SetAcceptsMessageType(typeId,not self:AcceptsMessageType(typeId))
end

-- @desc Sets all type id acceptance properties for registered typeids.
-- @arg accepts boolean Whether this frame accepts all (or no) typeIds.
-- @note Only all currently registered typeIds are set or unset.
function DreamChat.FrameClass.prototype:SetAcceptsAllMessageTypes(accepts)
	local k,_;
	for k,_ in pairs(DreamChat.messageTypes) do
		self:SetAcceptsMessageType(k,accepts);
	end
end
